package pl.przemelek.challenges;

public class PhilosophersDinnerToken {

    static volatile boolean start = false;
    static int FOOD = Integer.MAX_VALUE/100;

    static class Table {
        private Object lock = new Object();

        private int food = FOOD;

        public int takeOneBit() {
            synchronized (lock) {
                if (food>0) {
                    food--;
                    return 1;
                } else {
                    return 0;
                }
            }
        }

    }

    private Table table = new Table();

    static class Chopstick {
        boolean taken;

        public boolean pickUp() {
            if (!taken) {
                taken = true;
                return true;
            }
            return false;
        }

        public void putDown() {
            taken=false;
        }
    }


    static class Philosopher implements Runnable {
        private final Chopstick left;
        private final Chopstick right;
        private final Table table;
        int hadLeft = 0;
        int hadRight = 0;


        private int eaten;

        Philosopher(Table table, Chopstick left, Chopstick right) {
            this.table = table;
            this.left = left;
            this.right = right;
        }

        @Override
        public void run() {
            while (!start);
            boolean stop = false;
            while (!stop) {
                boolean haveLeft = false;
                boolean haveRight = false;
                synchronized (table) {
                    haveLeft = left.pickUp();
                    if (haveLeft) {
                        haveRight=right.pickUp();
                    }
                }
                if (haveLeft) hadLeft++;
                if (haveRight) hadRight++;
                if (haveLeft && haveRight) {
                    int bit = table.takeOneBit();
                    eaten+= bit;
                    stop = bit==0;
                }

                if (haveLeft || haveRight)
                synchronized (table) {
                    if (haveLeft) left.putDown();
                    if (haveRight) right.putDown();
                }
            }
        }
    }

    public static void main(String[] args) {
        for (int j=0; j<10; j++) {
            Table table = new Table();
            Chopstick[] chopsticks = new Chopstick[5];
            for (int i = 0; i < chopsticks.length; i++) chopsticks[i] = new Chopstick();
            Philosopher[] philosophers = new Philosopher[5];
            for (int i = 0; i < philosophers.length; i++)
                philosophers[i] = new Philosopher(table, chopsticks[i], chopsticks[(i + 1) % chopsticks.length]);
            for (Philosopher philosopher : philosophers) new Thread(philosopher).start();
            start = true;
            long s = System.currentTimeMillis();
            while (table.food != 0) {
                try {
                    Thread.sleep(10);
                } catch (InterruptedException ie) {
                }
                ;
            }
            System.out.println(System.currentTimeMillis() - s);
            double diff = 0;
            for (int i = 0; i < philosophers.length; i++) {
                double percent = philosophers[i].eaten * 1.0 / FOOD * 100;
                System.out.println(i + " " + percent + " " + philosophers[i].hadLeft * 1.0 / philosophers[i].hadRight);
                diff += Math.abs(20 - percent);
            }
            System.out.println(diff);
            System.out.println("------");
        }
    }

}
